#include "Facebook.h"
#include <PlayrixEngine.h>
#include <queue>
#include <boost/format.hpp>
#include "android/JniWrapper/JniWrapper.h"
#include "Core/App/AppEventsDispatcher.h"
#include "Singleton.h"
using namespace Jni;

void QueueImageCallback(bool success, const std::string& filename);

namespace FB 
{

JNI_DECLARE_CLASS(JniPlayrixFacebook, "com/playrix/lib/PlayrixFacebook")
JNI_DECLARE_CLASS(JniPlayrixFacebookFriend, "com/playrix/lib/PlayrixFacebook$Friend")
// JNI_DECLARE_CLASS(JniPlayrixFacebookInvitableFriend, "com/playrix/lib/PlayrixFacebook$InvitableFriend")
JNI_DECLARE_CLASS(JniPlayrixFacebookAppRequest, "com/playrix/lib/PlayrixFacebook$AppRequest")
JNI_DECLARE_CLASS(JniFacebookAppEventsConstants, "com/facebook/appevents/AppEventsConstants")
JNI_DECLARE_CLASS(JniBundle, "android/os/Bundle")

class State : public Singleton<State> {
	friend class Singleton<State>;
	State() = default;
public:
	int AddDeferredLinkRequest(DeferredLinkCallback deferredLinkCallback) {
		Assert(utils::IsMainThread());
		if (!deferredLinkCallback) {
			return 0;
		}

		while(!_deferredLinkRequestId || (_deferredLinkRequests.find(_deferredLinkRequestId) != _deferredLinkRequests.end())) {
			++_deferredLinkRequestId;
		}

		_deferredLinkRequests[_deferredLinkRequestId] = deferredLinkCallback;
		return _deferredLinkRequestId++;
	}

	DeferredLinkCallback GetDeferredLinkRequest(int requestId) {
		Assert(utils::IsMainThread());
		if (!requestId) {
			return DeferredLinkCallback();
		}
		auto found = _deferredLinkRequests.find(requestId);
		if (found == _deferredLinkRequests.end()) {
			Halt("Not deferred link found request id");
			return DeferredLinkCallback();
		}
		auto result = found->second;
		_deferredLinkRequests.erase(found);
		return result;
	}
public:
	CallbackType _callback;
	std::queue<Action::Type> _startupQueue;
	std::map<int, DeferredLinkCallback> _deferredLinkRequests;
	int _deferredLinkRequestId = 1;
	bool _waitingClearPermissions = false;
};

static void CallOrQueue(Action::Type action)
{
	auto &state = State::Instance();
	if (state._callback) {
		state._callback(action);
	}
	else {
		state._startupQueue.push(action);
	}
}

static JniBundle CreateBundle(const Parameters &params)
{
	JniBundle jbundle = JniBundle::Construct(JniInt(static_cast<int>(params.size())));
	Parameters::const_iterator i = params.begin(), e = params.end();
	for (; i != e; ++i) {
		jbundle.CallMethod<JniVoid>("putString", JniString(i->first), JniString(i->second));
	}
	return jbundle;
}

void SetTimeout(int timeoutInSeconds)
{
	JniPlayrixFacebook::CallStaticMethod<JniVoid>("setTimeout", JniInt(timeoutInSeconds));
}

void SetUserPicturesSize(int width, int height)
{
	Assert((width > 0) && (height > 0));
	std::string picturesField = (boost::format("picture.width(%1%).height(%2%)%3%") % width % height % ((height == width) ? ".type(square)" : "")).str();
	JniPlayrixFacebook::CallStaticMethod<JniVoid>("setUserPicturesField", JniString(picturesField));
}

std::string GetSdkVersion()
{
	return JniPlayrixFacebook::CallStaticMethod<JniString>("getSdkVersion").String();
}

void Initialize()
{
}

void InitWithCallback(CallbackType callback)
{
	Log::Info("FB: SDK version: " + GetSdkVersion());

	auto &state = State::Instance();
	state._callback = callback;
	while (!state._startupQueue.empty()) {
		state._callback(state._startupQueue.front());
		state._startupQueue.pop();
	}
}

void FetchDeferredDeepLink(DeferredLinkCallback deferredLinkCallback)
{
	int requestId = State::Instance().AddDeferredLinkRequest(deferredLinkCallback);
	JniPlayrixFacebook::CallStaticMethod<JniVoid>("processDeferredDeeplink", JniInt(requestId));
}

void SetLoginBehavior(LoginBehavior behavior)
{
	JniPlayrixFacebook::CallStaticMethod<JniVoid>("setLoginBehavior", JniInt((int)behavior));
}

void LogIn()
{
	JniPlayrixFacebook::CallStaticMethod<JniVoid>("login");
}

LoginError GetLastLoginError()
{
	return (LoginError)(int)JniPlayrixFacebook::CallStaticMethod<JniInt>("getLastLoginError");
}

static void LogInLike()
{
	JniPlayrixFacebook::CallStaticMethod<JniVoid>("loginLike");
}

void LogOut()
{
	JniPlayrixFacebook::CallStaticMethod<JniVoid>("logout");
}

bool IsGameLiked()
{
	return JniPlayrixFacebook::CallStaticMethod<JniBoolean>("isGameLiked");
}

bool RequestLikesInformation()
{
	if (!IsLoggedIn()) { return false; }
	if (!CanCheckLike()) {
		LogInLike();
		return true;
	}

	JniPlayrixFacebook::CallStaticMethod<JniVoid>("requestLikesInfo");
	return true;
}

bool IsLoggedIn()
{
	return JniPlayrixFacebook::CallStaticMethod<JniBoolean>("loggedIn");
}

bool CanPublish()
{
	return JniPlayrixFacebook::CallStaticMethod<JniBoolean>("canPublish");
}

bool CanCheckLike()
{
	return JniPlayrixFacebook::CallStaticMethod<JniBoolean>("canCheckLike");
}

bool AuthInProcess()
{
	return JniPlayrixFacebook::CallStaticMethod<JniBoolean>("authInProcess");
}

bool LoadingUserInfo()
{
	return JniPlayrixFacebook::CallStaticMethod<JniBoolean>("requestUserInProcess");
}

bool RequestUserInformation()
{
	if (!IsLoggedIn()) {
		return false;
	}

	JniPlayrixFacebook::CallStaticMethod<JniVoid>("requestUser");
	return true;
}

bool RequestFriendsInformation(int totalRetrieveLimit, RequestMode requestMode)
{
	if (!IsLoggedIn()) {
		return false;
	}

	return JniPlayrixFacebook::CallStaticMethod<JniBoolean>("requestFriends", JniInt(totalRetrieveLimit), JniInt((int)requestMode));
}

// bool RequestInvitableFriendsInformation(int totalRetrieveLimit, RequestMode requestMode)
// {
// 	if (!IsLoggedIn()) {
// 		return false;
// 	}

// 	return JniPlayrixFacebook::CallStaticMethod<JniBoolean>("requestInvitableFriends", JniInt(totalRetrieveLimit), JniInt((int)requestMode));
// }

bool IsWaitingForFriends() 
{
	return JniPlayrixFacebook::CallStaticMethod<JniBoolean>("requestFriendsInProcess");
}

void PostToOpenGraph(const std::string& actionWithPrefix, const std::string& objurl, const std::string& objname)
{
	JniPlayrixFacebook::CallStaticMethod<JniVoid>("postToOpenGraph", JniString(actionWithPrefix), JniString(objurl), JniString(objname));
}

void PostWithDialog(const Parameters &params)
{
	if (!IsLoggedIn()) {
		LogIn();
		return;
	}

	JniPlayrixFacebook::CallStaticMethod<JniVoid>("postWithDialog", CreateBundle(params));
}

void PostOnWall(const Parameters &params)
{
	JniPlayrixFacebook::CallStaticMethod<JniVoid>("postOnWall", CreateBundle(params));
}

Friend GetFriend(const std::string &fbId)
{
	JniPlayrixFacebookFriend jfriend = JniPlayrixFacebook::CallStaticMethod<JniPlayrixFacebookFriend>("friend", JniString(fbId));

	FB::Friend result;
	result.fbId = fbId;
	result.name = jfriend.CallMethod<JniString>("username").String();
	result.first_name = jfriend.CallMethod<JniString>("firstName").String();
	result.pic_url = jfriend.CallMethod<JniString>("pictureUrl").String();

	return result;
}

std::vector<std::string> FriendIds()
{
	if (!IsLoggedIn()) {
		return std::vector<std::string>();
	}

	return Helpers::MakeStringVector(JniPlayrixFacebook::CallStaticMethod<JniObjectArray>("friendIds"));
}

// InvitableFriend GetInvitableFriend(const std::string& inviteId)
// {
// 	JniPlayrixFacebookInvitableFriend jfriend = JniPlayrixFacebook::CallStaticMethod<JniPlayrixFacebookInvitableFriend>("invitableFriend", JniString(inviteId));

// 	FB::InvitableFriend result;
// 	result.inviteId = inviteId;
// 	result.name = jfriend.CallMethod<JniString>("username").String();
// 	result.pic_url = jfriend.CallMethod<JniString>("pictureUrl").String();

// 	//Log::Info("InvitableFriend = " + result.name); // Debug output
// 	return result;
// }

// std::vector<std::string> InviteIds()
// {
// 	if (!IsLoggedIn()) {
// 		return std::vector<std::string>();
// 	}

// 	return Helpers::MakeStringVector(JniPlayrixFacebook::CallStaticMethod<JniObjectArray>("invitableFriendIds"));
// }

std::string UserName()
{
	return JniPlayrixFacebook::CallStaticMethod<JniString>("getUsername").String();
}

std::string UserFirstName()
{
	return JniPlayrixFacebook::CallStaticMethod<JniString>("getUserFirstname").String();
}

std::string UserId()
{
	return JniPlayrixFacebook::CallStaticMethod<JniString>("getUserId").String();
}

std::string UserPicUrl()
{
	return JniPlayrixFacebook::CallStaticMethod<JniString>("getUserPicUrl").String();
}

void SendRequest(const Parameters &params)
{
	if (!IsLoggedIn()) {
		return;
	}

	CallOrQueue(Action::REQUEST_DIALOG_SHOWN);
	JniPlayrixFacebook::CallStaticMethod<JniVoid>("request", CreateBundle(params));
}

// void InviteRequest(const std::string& appLink) {
// 	if(!IsLoggedIn()) {
// 		Log::Error("InviteRequest before log in");
// 		return;
// 	}

// 	JniPlayrixFacebook::CallStaticMethod<JniVoid>("invite", JniString(appLink));
// }

bool RequestAppRequestsInformation(int totalRetrieveLimit, RequestMode requestMode)
{
	if (!IsLoggedIn()) {
		return false;
	}

	return JniPlayrixFacebook::CallStaticMethod<JniBoolean>("requestAppRequests", JniInt(totalRetrieveLimit), JniInt((int)requestMode));
}

std::vector<AppRequest> GetAppRequests()
{
	JniObjectArray jarray = JniPlayrixFacebook::CallStaticMethod<JniObjectArray>("getAppRequests");

	std::vector<AppRequest> result;
	size_t size = jarray.Size();
	result.reserve(size);
	for (size_t i = 0; i < size; ++i) {
		JniPlayrixFacebookAppRequest jreq = jarray.Get(i);

		AppRequest req;
		req.requestId = jreq.CallMethod<JniString>("getRequestId").String();
		req.senderId = jreq.CallMethod<JniString>("getSenderId").String();
		req.message = jreq.CallMethod<JniString>("getMessage").String();
		req.data = jreq.CallMethod<JniString>("getData").String();

		result.push_back(req);
	}

	return result;
}

void DeleteAppRequest(const std::string &requestId)
{
	if (!IsLoggedIn()) {
		return;
	}

	AppRequestsToDelete::Instance().DeleteAppRequest(requestId);
}

static void SendDeleteAppRequest(const std::string &requestId) {
	JniPlayrixFacebook::CallStaticMethod<JniVoid>("deleteAppRequest", JniString(requestId));
}

#if !defined(PRODUCTION)
bool ClearPermissions()
{
	if (!IsLoggedIn()) {
		return false;
	}

	if (State::Instance()._waitingClearPermissions) {
		return false;
	}

	State::Instance()._waitingClearPermissions = true;

	JniPlayrixFacebook::CallStaticMethod<JniVoid>("clearPermissions");
	return true;
}
#endif

static void logEvent(const std::string& event, Parameters params) {
	JniPlayrixFacebook::CallStaticMethod<JniVoid>("logEvent", JniString(event), CreateBundle(params));
}

void LogEventActivateApp()
{
	Log::Info("FB::LogEventActivateApp");
	
	JniPlayrixFacebook::CallStaticMethod<JniVoid>("logActivation");
}

void LogEventFBLogin()
{
	Log::Info("FB::LogEventFBLogin");

	std::string evtConstant = JniFacebookAppEventsConstants::GetStaticField<JniString>("EVENT_NAME_COMPLETED_REGISTRATION").String();
	logEvent(evtConstant, Parameters());
}

void LogEventCompletedTutorial()
{
	Log::Info("FB::LogEventCompletedTutorial");
	
	std::string evtConstant = JniFacebookAppEventsConstants::GetStaticField<JniString>("EVENT_NAME_COMPLETED_TUTORIAL").String();
	logEvent(evtConstant, Parameters());
}

void LogEventAchievedLevel(int level)
{
	Log::Info("FB::LogEventAchievedLevel " + utils::lexical_cast(level));

	std::string evtConstant = JniFacebookAppEventsConstants::GetStaticField<JniString>("EVENT_NAME_ACHIEVED_LEVEL").String();
	std::string paramName = JniFacebookAppEventsConstants::GetStaticField<JniString>("EVENT_PARAM_LEVEL").String();

	Parameters params;
	params.emplace(paramName, utils::lexical_cast(level));
	logEvent(evtConstant, params);
}

void LogEventPurchased(float value, const std::string& currency)
{
	Log::Info("FB::LogEventPurchased " + utils::lexical_cast(value) + " " + currency);

	JniPlayrixFacebook::CallStaticMethod<JniVoid>("logPurchase", JniFloat(value), JniString(currency));
}

void LogEventAchievement(const std::string& achievement)
{
	Log::Info("FB::LogEventAchievement " + achievement);

	std::string evtConstant = JniFacebookAppEventsConstants::GetStaticField<JniString>("EVENT_NAME_UNLOCKED_ACHIEVEMENT").String();
	std::string paramName = JniFacebookAppEventsConstants::GetStaticField<JniString>("EVENT_PARAM_DESCRIPTION").String();

	Parameters params;
	params.emplace(paramName, achievement);
	logEvent(evtConstant, params);
}

void LogEventCustom(const std::string& event, const FB::Parameters& parameters)
{
	Log::Info("FB::LogEventCustom " + event);

	logEvent(event, parameters);
}

}

JNI_DECLARE_METHOD(JniVoid, com_playrix_lib_PlayrixFacebook, nativeOnLoginCompleted)
	CallOrQueue(FB::Action::LOGIN_COMPLETED);
}

JNI_DECLARE_METHOD(JniVoid, com_playrix_lib_PlayrixFacebook, nativeOnLoginCanceled)
	CallOrQueue(FB::Action::LOGIN_CANCELED);
}

JNI_DECLARE_METHOD(JniVoid, com_playrix_lib_PlayrixFacebook, nativeOnLoginFailed)
	CallOrQueue(FB::Action::LOGIN_FAILED);
}

JNI_DECLARE_METHOD(JniVoid, com_playrix_lib_PlayrixFacebook, nativeOnLoginForPublishCompleted)
	CallOrQueue(FB::Action::LOGIN_FOR_PUBLISH_COMPLETED);
}

JNI_DECLARE_METHOD(JniVoid, com_playrix_lib_PlayrixFacebook, nativeOnLoginForPublishFailed)
	CallOrQueue(FB::Action::LOGIN_FOR_PUBLISH_FAILED);
}

JNI_DECLARE_METHOD(JniVoid, com_playrix_lib_PlayrixFacebook, nativeOnLoginForPublishCanceled)
	CallOrQueue(FB::Action::LOGIN_FOR_PUBLISH_CANCELED);
}

JNI_DECLARE_METHOD(JniVoid, com_playrix_lib_PlayrixFacebook, nativeOnLogout)
	FB::AppRequestsToDelete::Instance().Clear();
	CallOrQueue(FB::Action::LOGOUT);
}

JNI_DECLARE_METHOD(JniVoid, com_playrix_lib_PlayrixFacebook, nativeOnUserInfoAvailable)
	CallOrQueue(FB::Action::USERINFO_AVAILABLE);
}

JNI_DECLARE_METHOD(JniVoid, com_playrix_lib_PlayrixFacebook, nativeOnUserInfoFailed)
	CallOrQueue(FB::Action::USERINFO_FAILED);
}

JNI_DECLARE_METHOD(JniVoid, com_playrix_lib_PlayrixFacebook, nativeOnFriendsInfoAvailable)
	CallOrQueue(FB::Action::FRIENDSINFO_AVAILABLE);
}

JNI_DECLARE_METHOD(JniVoid, com_playrix_lib_PlayrixFacebook, nativeOnFriendsInfoFailed)
	CallOrQueue(FB::Action::FRIENDSINFO_FAILED);
}

// JNI_DECLARE_METHOD(JniVoid, com_playrix_lib_PlayrixFacebook, nativeOnInvitableFriendsInfoAvailable)
// 	CallOrQueue(FB::Action::INVITABLEFRIENDS_AVAILABLE);
// }

// JNI_DECLARE_METHOD(JniVoid, com_playrix_lib_PlayrixFacebook, nativeOnInvitableFriendsInfoFailed)
// 	CallOrQueue(FB::Action::INVITABLEFRIENDS_FAILED);
// }

JNI_DECLARE_METHOD(JniVoid, com_playrix_lib_PlayrixFacebook, nativeOnDialogPosted)
	CallOrQueue(FB::Action::DIALOG_POSTED);
}

JNI_DECLARE_METHOD(JniVoid, com_playrix_lib_PlayrixFacebook, nativeOnDialogCanceled)
	CallOrQueue(FB::Action::DIALOG_CANCELED);
}

JNI_DECLARE_METHOD(JniVoid, com_playrix_lib_PlayrixFacebook, nativeOnRequestSent)
	CallOrQueue(FB::Action::REQUEST_SENT);
}

JNI_DECLARE_METHOD(JniVoid, com_playrix_lib_PlayrixFacebook, nativeOnRequestSentError)
	CallOrQueue(FB::Action::REQUEST_SENT_ERROR);
}

JNI_DECLARE_METHOD(JniVoid, com_playrix_lib_PlayrixFacebook, nativeOnRequestFailed)
	CallOrQueue(FB::Action::REQUEST_FAILED);
}

JNI_DECLARE_METHOD(JniVoid, com_playrix_lib_PlayrixFacebook, nativeOnRequestDialogClosed)
	CallOrQueue(FB::Action::REQUEST_DIALOG_CLOSED);
}

// JNI_DECLARE_METHOD(JniVoid, com_playrix_lib_PlayrixFacebook, nativeOnInviteSent)
// 	CallOrQueue(FB::Action::INVITE_SENT);
// }

// JNI_DECLARE_METHOD(JniVoid, com_playrix_lib_PlayrixFacebook, nativeOnInviteCanceled)
// 	CallOrQueue(FB::Action::INVITE_SENT_CANCEL);
// }

// JNI_DECLARE_METHOD(JniVoid, com_playrix_lib_PlayrixFacebook, nativeOnInviteError)
// 	CallOrQueue(FB::Action::INVITE_SENT_ERROR);
// }

JNI_DECLARE_METHOD(JniVoid, com_playrix_lib_PlayrixFacebook, nativeOnGetAppRequestsInfoAvailable)
	CallOrQueue(FB::Action::APPREQUEST_AVAILABLE);
}

JNI_DECLARE_METHOD(JniVoid, com_playrix_lib_PlayrixFacebook, nativeOnGetAppRequestsInfoFailed)
	CallOrQueue(FB::Action::APPREQUEST_FAILED);
}

JNI_DECLARE_METHOD(JniVoid, com_playrix_lib_PlayrixFacebook, nativeOnLikesInfoAvailable)
	CallOrQueue(FB::Action::LIKESINFO_AVAILABLE);
}

JNI_DECLARE_METHOD(JniVoid, com_playrix_lib_PlayrixFacebook, nativeOnLikesInfoFailed)
	CallOrQueue(FB::Action::LIKESINFO_FAILED);
}

JNI_DECLARE_METHOD(JniVoid, com_playrix_lib_PlayrixFacebook, nativeOnClearPermissionsSucceeded)
	FB::State::Instance()._waitingClearPermissions = false;
	CallOrQueue(FB::Action::CLEAR_PERMISSIONS_SUCCEEDED);
	FB::LogOut();
}

JNI_DECLARE_METHOD(JniVoid, com_playrix_lib_PlayrixFacebook, nativeOnClearPermissionsInfoFailed)
	FB::State::Instance()._waitingClearPermissions = false;
	CallOrQueue(FB::Action::CLEAR_PERMISSIONS_FAILED);
}

JNI_DECLARE_METHOD(JniVoid, com_playrix_lib_PlayrixFacebook, nativeOnDeleteAppRequest, JniBoolean, succeeded, JniString, appRequestId)
	FB::AppRequestsToDelete::Instance().OnAppRequestDelete(succeeded, appRequestId.String());
}

JNI_DECLARE_METHOD(JniVoid, com_playrix_lib_PlayrixFacebook, nativeOnDeferredAppLinkDataFetched, JniInt, requestId, JniString, url)

	auto callback = FB::State::Instance().GetDeferredLinkRequest(requestId);
	auto urlString = url.String();

	if (callback) {
		callback(urlString);
		return;
	}

	if (urlString.empty()) {
		return;
	}

	if (Core::AppEventsDispatcher::Instance().Call().OnOpenUrl(urlString, nullptr, nullptr, std::string(), nullptr)) {
		return;
	}
	Core::appDelegate->OnUrlActivate(urlString, std::string());
}
