#include "types.h"

#include <HelpShift.h>

#include "android/JniWrapper/JniWrapper.h"
#include "Core/App/AppEventsDispatcher.h"
#include "Singleton.h"

#include <json/json.h>

using namespace Jni;

namespace HelpShift
{
	JNI_DECLARE_CLASS(JniPlayrixHelpshift, "com/playrix/lib/HelpshiftWrapper")

	class State : public Singleton<State> {
		friend class Singleton<State>;
		State() = default;
	public:
		std::vector <std::pair <std::string, std::string> > _helpshiftFaqSections;
		std::vector<std::string> _tags;
		std::vector<std::string> _info;
		std::map<std::string, std::string> _metadata;
		std::string _loginId;
		std::string _name;
		std::string _logsUrl;
		IHelpshiftDelegate* _delegate = nullptr;
		bool _isEnabled = true;
		int _notificationCount = 0;
		std::string _pendingPushToken;

		void ApplyPendingPushToken() {
			if (!_pendingPushToken.empty()) {
				JniPlayrixHelpshift::CallStaticMethod<JniVoid>("registerPushToken", JniString(_pendingPushToken));
				_pendingPushToken = "";
			}
		}
	};

	void SendCampaigns(const std::unordered_map<std::string, VariableType>& campaigns)
	{
		if (!campaigns.empty())
		{
			for (const auto & iter : campaigns)
			{
				switch (iter.second.type())
				{
					case VariableType::ValueTypes::String:
						JniPlayrixHelpshift::CallStaticMethod<JniVoid>("addStringCampaignProp", JniString(iter.first), JniString(iter.second.value<std::string>()));
						break;

					case VariableType::ValueTypes::Integer:
						JniPlayrixHelpshift::CallStaticMethod<JniVoid>("addIntegerCampaignProp", JniString(iter.first), JniInt(iter.second.value<int>()));
						break;

					case VariableType::ValueTypes::Boolean:
						JniPlayrixHelpshift::CallStaticMethod<JniVoid>("addBooleanCampaignProp", JniString(iter.first), JniBoolean(iter.second.value<bool>()));
						break;

					case VariableType::ValueTypes::DateTime:
						JniPlayrixHelpshift::CallStaticMethod<JniVoid>("addDateCampaignProp", JniString(iter.first), JniLong(iter.second.value<time_t>()));
						break;
				}
			}
		}
	}

	static void UpdatePlayerInfo(IHelpshiftDelegate& delegate)
	{
		auto &state = State::Instance();
		state._tags.clear();
		state._info.clear();
		state._metadata.clear();
		delegate.OnGetPlayerInfo(state._loginId, state._name, state._tags, state._info, state._logsUrl, state._metadata);
		state._metadata["Hockey Search URL"] = state._logsUrl.empty() ? "empty" : state._logsUrl;

		bool isLogged = HelpShift::JniPlayrixHelpshift::CallStaticMethod<JniBoolean>("setHSData",
			JniString(state._loginId), JniString(state._name), Helpers::MakeStringArray(state._tags), Helpers::MakeStringArray(state._info), Helpers::MakeStringMap(state._metadata));
		if (isLogged)
		{
			std::unordered_map<std::string, VariableType> campaigns;
			delegate.OnGetCampaignsProperties(campaigns);
			SendCampaigns(campaigns);
		}
	}

	static void UpdateEnableContactUs(IHelpshiftDelegate& delegate) {
		JniPlayrixHelpshift::CallStaticMethod<JniVoid>("setEnableContactUs", JniInt(static_cast<int>(delegate.EnableShowContactUs())));
	}

	//добавляет пункт меню в массив для динамической формы главного меню faq
	static void AddFlowToArray(IHelpshiftDelegate& delegate, Json::Value &array, const std::string& id, const std::string& name, const std::string& extratags = "")
	{
		//проверим параметры
		if (id.empty() || name.empty()) {
			Assert(false);
			return;
		}

		Json::Value data;
		data["id"] = id;
		data["name"] = name + (extratags.empty() ? "" : ("," + extratags));
		data["text"] = delegate.GetTextByID("Helpshift_" + name);

		//добавим его в массив
		array.append(data);
	};

	class HelpshiftWrapper::AppEventsDelegate : public Core::AppEventsDispatcher::Delegate {
	public:
		virtual bool OnOpenUrl(const std::string &urlString, ApplicationContext *application, NSURL *url, const std::string &sourceApplication, id annotation) override {
			if (!_wrapper) {
				return false;
			}
			return _wrapper->OpenUrl(urlString);
		}

		virtual void OnDidRegisterForRemoteNotifications(ApplicationContext *application, NotificationToken deviceToken) override {
			if (!deviceToken.empty()) {
				State::Instance()._pendingPushToken = deviceToken;
				if(_wrapper && _wrapper->IsActive()) {
					State::Instance().ApplyPendingPushToken();
				}
			}
		}
	public:
		HelpshiftWrapper *_wrapper = nullptr;
	};

	HelpshiftWrapper::AppEventsDelegate *HelpshiftWrapper::GetAppEventsDelegate() {
		static AppEventsDelegate *appEventsDelegate = []() {
			AppEventsDelegate *result = new AppEventsDelegate();
			Core::AppEventsDispatcher::Instance().Subscribe(result);
			return result;
		}();
		return appEventsDelegate;
	}

	bool HelpshiftWrapper::_isInitialized = false;

	void HelpshiftWrapper::Initialize() {
		InitializeWithParameters(std::string(), std::string(), std::string());
	}

	void HelpshiftWrapper::InitializeWithParameters(const std::string& appId, const std::string& apiKey, const std::string& domain) {
		if (_isInitialized) {
			Halt("Try initialize helpshift again");
			return;
		}

		_isInitialized = JniPlayrixHelpshift::CallStaticMethod<JniBoolean>("isInitialized");
		Assert2(_isInitialized, "Helpshift must be initialized from Application::onCreate");

		Log::Info("Helpshift SDK version: " + GetSdkVersion());

		GetAppEventsDelegate();
	}

	HelpshiftWrapper::HelpshiftWrapper(const std::string& urlPrefix, IHelpshiftDelegate& delegate)
		: _urlPrefix(urlPrefix)
		, _delegate(delegate)
	{
		State::Instance()._delegate = &_delegate;
		GetAppEventsDelegate()->_wrapper = this;
		State::Instance().ApplyPendingPushToken();
	}

	HelpshiftWrapper::~HelpshiftWrapper() {
		if (State::Instance()._delegate == &_delegate) {
			State::Instance()._delegate = nullptr;
		}
		auto appEventsDelegate = GetAppEventsDelegate();
		if (appEventsDelegate->_wrapper == this) {
			appEventsDelegate->_wrapper = nullptr;
		}
	}

	std::string HelpshiftWrapper::GetSdkVersion() {
		return JniPlayrixHelpshift::CallStaticMethod<JniString>("getSdkVersion").String();
	}

	void HelpshiftWrapper::SetEnabled(bool enable)
	{
		_isEnabled = enable;
		State::Instance()._isEnabled = enable;
		JniPlayrixHelpshift::CallStaticMethod<JniVoid>("setEnabled", JniBoolean(enable));
	}

	void HelpshiftWrapper::SetFaqSections(const std::vector <std::pair <std::string, std::string> > & faqSections)
	{
		if (!IsActive()) {
			return;
		}

		State::Instance()._helpshiftFaqSections = faqSections;
	}

	void HelpshiftWrapper::Logout()
	{
		if (!IsActive()) {
			return;
		}

		JniPlayrixHelpshift::CallStaticMethod<JniVoid>("logout");
	}

	void HelpshiftWrapper::SetLanguage(const std::string& language)
	{
		if (!IsActive()) {
			return;
		}

		JniPlayrixHelpshift::CallStaticMethod<JniVoid>("setLanguage", JniString(language));
	}

	void HelpshiftWrapper::EnableShowConversationResolutionQuestion(bool enabled)
	{
		if (!IsActive()) {
			return;
		}

		JniPlayrixHelpshift::CallStaticMethod<JniVoid>("setShowConversationResolutionQuestion", JniBoolean(enabled));
	}

	void HelpshiftWrapper::EnableShowSearchOnNewConversation(bool enabled)
	{
		if (!IsActive()) {
			return;
		}

		JniPlayrixHelpshift::CallStaticMethod<JniVoid>("setShowSearchOnNewConversation", JniBoolean(enabled));
	}

	void HelpshiftWrapper::EnableShowConversationInfoScreen(bool enabled)
	{
		if (!IsActive()) {
			return;
		}

		JniPlayrixHelpshift::CallStaticMethod<JniVoid>("setShowConversationInfoScreen", JniBoolean(enabled));
	}

	void HelpshiftWrapper::EnableGotoConversationAfterContactUs(bool enabled)
	{
		if (!IsActive())
		{
			return;
		}

		JniPlayrixHelpshift::CallStaticMethod<JniVoid>("setGotoConversationAfterContactUs", JniBoolean(enabled));
	}

	void HelpshiftWrapper::ShowConversation()
	{
		if (!IsActive()) {
			return;
		}

		UpdatePlayerInfo(_delegate);

		JniPlayrixHelpshift::CallStaticMethod<JniVoid>("showConversation");
	}

	void HelpshiftWrapper::ShowFaq()
	{
		if (!IsActive()) {
			return;
		}

		UpdatePlayerInfo(_delegate);
		UpdateEnableContactUs(_delegate);
		Json::Value flowsMainMenu = Json::Value(Json::arrayValue);
		for(const auto &item : State::Instance()._helpshiftFaqSections) {
			AddFlowToArray(_delegate, flowsMainMenu, item.first, item.second);
		}

		std::vector <std::pair <std::string, std::string> > contactUs;
		_delegate.OnGetCustomContacUsFlow(contactUs);
		Json::Value flowsContactUs = Json::Value(Json::arrayValue);
		for(const auto &item : contactUs) {
			Json::Value data;
			data["tag"] = item.second;
			data["text"] = item.first;
			flowsContactUs.append(data);
		}

		JniPlayrixHelpshift::CallStaticMethod<JniVoid>("showFaq",
													   JniString(Json::FastWriter().write(flowsMainMenu)),
													   JniString(Json::FastWriter().write(flowsContactUs)));
	}

	bool HelpshiftWrapper::ShowFaqSection(const std::string& sectionId)
	{
		if (!IsActive()) {
			return false;
		}

		UpdatePlayerInfo(_delegate);
		UpdateEnableContactUs(_delegate);
		JniPlayrixHelpshift::CallStaticMethod<JniVoid>("showFaqSection", JniString(sectionId));
		return true;
	}

	bool HelpshiftWrapper::ShowSingleFaq(const std::string& sectionId, const std::string& questionId)
	{
		if (!IsActive()) {
			return false;
		}

		UpdatePlayerInfo(_delegate);
		UpdateEnableContactUs(_delegate);
		JniPlayrixHelpshift::CallStaticMethod<JniVoid>("showSingleFaq", JniString(sectionId), JniString(questionId));
		return true;
	}

	bool HelpshiftWrapper::IsFAQShown() const
	{
		if (!_isInitialized) {
			return false;
		}

		return JniPlayrixHelpshift::CallStaticMethod<JniBoolean>("isFAQShown");
	}

	int HelpshiftWrapper::GetNotificationsCount() const
	{
		if (!IsActive()) {
			return 0;
		}
		return State::Instance()._notificationCount;
	}

	void HelpshiftWrapper::UpdateNotificationCount()
	{
		if (!IsActive()) {
			return;
		}
		UpdatePlayerInfo(_delegate);
		JniPlayrixHelpshift::CallStaticMethod<JniVoid>("getNotificationCount");
	}
};

JNI_DECLARE_METHOD(JniVoid, com_playrix_lib_HelpshiftWrapper, nativeOnNewConversationStarted)
	const auto &state = HelpShift::State::Instance();
	if (state._delegate && state._isEnabled) {
		state._delegate->OnNewConversationStarted(state._logsUrl);
	}
}

JNI_DECLARE_METHOD(JniVoid, com_playrix_lib_HelpshiftWrapper, nativeOnReceiveNotification, JniInt, count)
	auto &state = HelpShift::State::Instance();
	state._notificationCount = count;
	if (state._delegate && state._isEnabled) {
		state._delegate->OnChangedNotificationsCount(count);
	}
}

JNI_DECLARE_METHOD(JniVoid, com_playrix_lib_HelpshiftWrapper, nativeOnUserCompletedCustomerSatisfactionSurvey, JniInt, rating, JniString, feedback)
	auto &state = HelpShift::State::Instance();
	if (state._delegate && state._isEnabled) {
		state._delegate->OnUserCompletedCustomerSatisfactionSurvey(rating, feedback.String());
	}
}